unit EraseThread;

interface

{.$DEFINE SIMULATION_MODE}

uses
   Windows, Messages, SysUtils, Classes
   {$IFDEF SIMULATION_MODE}
   , Forms, StdCtrls, Graphics
   {$ENDIF}
   ;

type
  TEraseMethod = class;
  TEraseCyleType = (ecZero, ecOne, ecRandom);

  TEraseThread = class(TThread)
  private
    DeviceID : String;
    FTotalBytes : Int64;
    FBytesPerSector : DWORD;
    FBytesLeft : Int64;
    BlockSize : DWORD;
    Buffer : Pointer;
    Handle : THandle;
    CleanUpDone : Boolean;
    FRunning : Boolean;
    FCurrentCycle : Byte;
    FEraseMethod : TEraseMethod;
    FStartTime : TDateTime;
  protected
    SectorsPerBlock : DWORD;
    procedure CleanUp;
    procedure DoEraseCycle(cycleType : TEraseCyleType);
    procedure FillBufferWithRandom;
    function GetCycle(Index : Byte) : TEraseCyleType;
    function GetCycleCount : Integer;
  public
    constructor Create(ADeviceID : String; AEraseMethod : TEraseMethod; BufferSize : Integer);
    destructor Destroy; override;
    procedure Execute; override;
    property BytesPerSector : DWORD read FBytesPerSector;
    property TotalBytes : Int64 read FTotalBytes;
    property BytesLeft : Int64 read FBytesLeft;
    property Running : Boolean read FRunning;
    property Cycles[Index : Byte] : TEraseCyleType read GetCycle;
    property CyclesCount : Integer read GetCycleCount;
    property CurrentCycle : Byte read FCurrentCycle;
    property StartTime : TDateTime read FStartTime;
  end;


  TEraseMethod = class
  protected
    FCycleCount : Integer;
  public
    Cycles : Array of TEraseCyleType;
    constructor Create(const a : Array of TEraseCyleType);
    property CycleCount : Integer read FCycleCount;
  end;

procedure ShowSimulationInfo(Form : TObject);

implementation

Uses
  JwaWinIoctl, JwaWinBase, DiskTools;

const
  GENERIC_READWRITE    = GENERIC_READ or GENERIC_WRITE;
  FILE_SHARE_READWRITE = FILE_SHARE_READ or FILE_SHARE_WRITE;
  SFinished = 'Erasing successfully completed';
  SAborted = 'Erasing of disk "%s" has been stopped by the user';

procedure ShowSimulationInfo(Form : TObject);
{$IFNDEF SIMULATION_MODE}
begin
end;
{$ELSEIF true}
Var
  f : TForm;
begin
  f:=TForm(form);
  f.Color:=clLime;
  f.Caption:=f.Caption+' [simulation mode]';
end;
{$IFEND}

constructor TEraseMethod.Create(const a : Array of TEraseCyleType);
Var
  i,len : Integer;
begin
  len:=High(a)-Low(a)+1;
  FCycleCount:=len;
  SetLength(cycles,len);
  For I:=0 TO len-1 do cycles[i]:=a[i];
end;

constructor TEraseThread.Create(ADeviceID : String; AEraseMethod : TEraseMethod; BufferSize : Integer);
Var
  discGeometry : DISK_GEOMETRY_EX;
  bout : DWORD;
  res : LongBool;
begin
  inherited Create(false);
  Randomize;
  FRunning:=False;
  DeviceID:=ADeviceID;
  CleanUpDone:=False;
  FEraseMethod:=AEraseMethod;
  FreeOnTerminate:=False;
  Handle:=Windows.CreateFile(PChar(DeviceID),GENERIC_READ or GENERIC_WRITE, 0, nil,
       OPEN_EXISTING ,FILE_FLAG_WRITE_THROUGH or FILE_FLAG_NO_BUFFERING,0);
  IF (Handle = INVALID_HANDLE_VALUE) Then RaiseLastOSError;

  res:=Windows.DeviceIOControl(Handle,IOCTL_DISK_GET_DRIVE_GEOMETRY_EX,nil,0,@discGeometry,sizeOf(discGeometry),bout,nil);
  if (NOT res) then begin
    CloseHandle(Handle);
    RaiseLastOSError;
  end;
  FBytesPerSector:=discGeometry.Geometry.BytesPerSector;
  FTotalBytes:=discGeometry.DiskSize.QuadPart;

  SectorsPerBlock:=16 * (1 shl BufferSize);
  BlockSize:=BytesPerSector * SectorsPerBlock;
  //MessageBox(0,PChar(Format('BlockSize: %d',[BlockSize])),'',MB_OK);
  Buffer:=VirtualAlloc(nil,BlockSize,MEM_RESERVE or MEM_COMMIT,PAGE_READWRITE);
  FCurrentCycle:=1;
end;

function TEraseThread.GetCycle(Index : Byte) : TEraseCyleType;
begin
  Result:=FEraseMethod.cycles[Index];
end;

function TEraseThread.GetCycleCount : Integer;
begin
  Result:=FEraseMethod.CycleCount;
end;

destructor TEraseThread.Destroy;
begin
  CleanUp;
  inherited Destroy;
end;

procedure TEraseThread.CleanUp;
begin
  If CleanUpDone Then exit;
  try
    CloseHandle(Handle);
    VirtualFree(Buffer,0,MEM_RELEASE);
    FRunning:=False;
  finally
    CleanUpDone:=true;
  end;
end;

procedure TEraseThread.Execute;
Var
  I : Integer;
  bout : DWORD;
begin
  FRunning:=True;
  FStartTime:=Now;
  try
    //Deleting all partitions (only required on Windows Vista)
    Windows.DeviceIOControl(Handle,IOCTL_DISK_DELETE_DRIVE_LAYOUT,nil,0,nil,0,bout,nil);
    For I:=Low(FEraseMethod.cycles) TO High(FEraseMethod.cycles) DO begin
      DoEraseCycle(FEraseMethod.cycles[I]);
      inc(FCurrentCycle);
    end;
    If Not terminated Then begin
      MessageBox(0,PChar(SFinished),'Task completed',MB_OK or MB_ICONINFORMATION);
      MessageBeep(0);
    end else
      MessageBox(0,PChar(Format(SAborted,[DeviceID])),'Task aborted',
        MB_OK or MB_ICONSTOP);
  except
    on E: Exception do
      MessageBox(0,PChar(e.Message),'Error occured while erasing disk',MB_OK or MB_ICONERROR);
  end;
  CleanUp;
  FRunning:=False;
end;

procedure TEraseThread.DoEraseCycle(cycleType : TEraseCyleType);
Var
  bw : DWORD;
  B : Boolean;
begin
  FBytesLeft:=FTotalBytes;
  //MessageBox(0,PChar(Format('Cycle: %d',[Integer(cycleType)])),'',MB_OK);
  case cycleType of
    ecRandom : FillBufferWithRandom;
    ecZero   : FillChar(Buffer^,BlockSize,0);
    ecOne    : FillChar(Buffer^,BlockSize,255);
  end;
  if (SetFilePointer(Handle,FILE_BEGIN,nil,0) <> 0) Then RaiseLastOSError;
  While (FBytesLeft>=BlockSize) AND Not Terminated do begin
    {$IFNDEF SIMULATION_MODE}
      if not Windows.WriteFile(Handle,buffer^,BlockSize,bw,nil) Then RaiseLastOSError;
      if (bw <> BlockSize) then
        Exception.CreateFmt('Unable to overwrite: writing of %d bytes in current block failed',[FBytesLeft]);
        If (cycleType = ecRandom) Then FillBufferWithRandom;
    {$ELSEIF true}
      bw:=BlockSize;
      sleep(1);
    {$IFEND}
    FBytesLeft:=FBytesLeft-bw;
  end;

  If (FBytesLeft>0) AND NOT Terminated Then begin
    {$IFNDEF SIMULATION_MODE}
      If NOT Windows.WriteFile(Handle,buffer^,FBytesLeft,bw,nil) Then RaiseLastOSError;
      if (bw <> FBytesLeft) then
        Exception.CreateFmt('Unable to overwrite: writing of %d bytes in current block failed',[FBytesLeft]);
    {$ELSEIF true}
      bw:=FBytesLeft;
    {$IFEND}
    FBytesLeft:=FBytesLeft-bw;
    if (FBytesLeft>0) Then raise Exception.CreateFmt('Failed to overwrite the last %d bytes',[FBytesLeft]);
  end;

  // Just to make sure that there are no bytes left which are not recognized by the OS
  repeat
    {$IFNDEF SIMULATION_MODE}
      B:=Windows.WriteFile(Handle,buffer^,FBytesPerSector,bw,nil);
    {$ELSEIF true}
      B:=False;
    {$IFEND}
  until ((NOT B) OR Terminated);
end;

procedure TEraseThread.FillBufferWithRandom;
Var
  n,i,r : Integer;
  pw : PDWORD;
  pb : PBYTE;
begin
  n:=BlockSize div (sizeOf(DWORD));
  r:=BlockSize mod (sizeOf(DWORD));
  pw:=buffer;
  For i:=1 To n do begin
    pw^:=DWORD(random($ffffffff));
    inc(pw);
  end;
  pb:=Pointer(pw);
  For i:=1 To r do begin
    pb^:=Byte(random($ff));
    inc(pb);
  end;
end;

end.
