unit IsAdmin;

interface

uses
  SysUtils, Windows, JwaWinBase, JwaWinNT;

function IsCurrentUserLocalAdministrator() : Boolean;

implementation

(*-------------------------------------------------------------------------

Delphi translation if MSKB 118626
http://support.microsoft.com/kb/118626/en

-
IsCurrentUserLocalAdministrator ()

This function checks the token of the calling thread to see if the caller
belongs to the Administrators group.

Return Value:
   TRUE if the caller is an administrator on the local machine.
   Otherwise, FALSE.
--------------------------------------------------------------------------*

/
*)

function IsCurrentUserLocalAdministrator() : Boolean;
Var
   dwStatus : DWORD;
   dwAccessMask : DWORD;
   dwAccessDesired : DWORD;
   dwACLSize : DWORD;
   dwStructureSize : DWORD;
   pACL_ : PACL;
   psidAdmin : PSID;

   Token : THandle;
   ImpersonationToken : THandle;

   ps : PRIVILEGE_SET;
   GenericMapping : GENERIC_MAPPING;

   psdAdmin : PSECURITY_DESCRIPTOR;
   SystemSidAuthority : SID_IDENTIFIER_AUTHORITY;

   res : LongBool;

(*
Determine if the current thread is running as a user that is a memberof
the local admins group.  To do this, create a security descriptor that
has a DACL which has an ACE that allows only local aministrators access.
Then, call AccessCheck with the current thread's token and the security
descriptor.  It will say whether the user could access an object if it
had that security descriptor.  Note: you do not need to actually create
the object.  Just checking access against the security descriptor alone
will be sufficient.
*)

const ACCESS_READ  : DWORD = 1;
const ACCESS_WRITE : DWORD = 2;


begin
  Result:=False;
  dwStructureSize := sizeof(PRIVILEGE_SET);
  SystemSidAuthority:=SECURITY_NT_AUTHORITY;
  psdAdmin:=nil;
  pACL_:=nil;
  try
    (*
      AccessCheck() requires an impersonation token.  We first get a primary
      token and then create a duplicate impersonation token.  The
      impersonation token is not actually assigned to the thread, but is
      used in the call to AccessCheck.  Thus, this function itself never
      impersonates, but does use the identity of the thread.  If the thread
      was impersonating already, this function uses that impersonation context.
    *)
    if ( not OpenThreadToken(GetCurrentThread(), TOKEN_DUPLICATE or TOKEN_QUERY,true, Token)) Then begin
      if (GetLastError() <> ERROR_NO_TOKEN) Then exit;
    if (not OpenProcessToken(GetCurrentProcess(),TOKEN_DUPLICATE or TOKEN_QUERY, Token)) Then
       exit;
  end;
  if (not DuplicateToken (Token, SecurityImpersonation, @ImpersonationToken)) Then exit;
    (*
      Create the binary representation of the well-known SID that
      represents the local administrators group.  Then create the
      security descriptor and DACL with an ACE that allows only local
      admins access.
      After that, perform the access check.  This will determine whether
      the current user is a local admin.
    *)
    if (not AllocateAndInitializeSid(@SystemSidAuthority, 2,
                                    SECURITY_BUILTIN_DOMAIN_RID,
                                    DOMAIN_ALIAS_RID_ADMINS,
                                    0, 0, 0, 0, 0, 0, psidAdmin)) Then exit;
      psdAdmin := PSECURITY_DESCRIPTOR(LocalAlloc(LPTR, SECURITY_DESCRIPTOR_MIN_LENGTH));
      if (psdAdmin = nil) Then exit;

      if (not InitializeSecurityDescriptor(psdAdmin, SECURITY_DESCRIPTOR_REVISION)) Then exit;

      // Compute size needed for the ACL.
      dwACLSize := sizeof(ACL) + sizeof(ACCESS_ALLOWED_ACE) +
                  GetLengthSid(psidAdmin) - sizeof(DWORD);

      pACL_ := PACL(LocalAlloc(LPTR, dwACLSize));
      if (pACL_ = nil) Then exit;

      if (not InitializeAcl(pACL_, dwACLSize, ACL_REVISION2)) Then exit;
      dwAccessMask := ACCESS_READ or ACCESS_WRITE;

      if (not AddAccessAllowedAce(pACL_, ACL_REVISION2, dwAccessMask,psidAdmin)) Then exit;
      if (not SetSecurityDescriptorDacl(psdAdmin, TRUE, pACL_, FALSE)) Then exit;

      (*
         AccessCheck validates a security descriptor somewhat; set the group
         and owner so that enough of the security descriptor is filled out to
         make AccessCheck happy.
      *)
      SetSecurityDescriptorGroup(psdAdmin, psidAdmin, FALSE);
      SetSecurityDescriptorOwner(psdAdmin, psidAdmin, FALSE);


      if (not IsValidSecurityDescriptor(psdAdmin)) Then exit;


      dwAccessDesired := ACCESS_READ;

      (*
         Initialize GenericMapping structure even though you
         do not use generic rights.
      *)
      GenericMapping.GenericRead    := ACCESS_READ;
      GenericMapping.GenericWrite   := ACCESS_WRITE;
      GenericMapping.GenericExecute := 0;
      GenericMapping.GenericAll     := ACCESS_READ or ACCESS_WRITE;

      res:=False;
      if (not AccessCheck(psdAdmin, ImpersonationToken, dwAccessDesired,
                       &GenericMapping, &ps, &dwStructureSize, &dwStatus,
                       res)) Then begin
         Result:=False;
         exit;
      end;
      Result:=res;
   finally
      // Clean up.
      if Assigned(pACL_) Then LocalFree(THandle(pACL_));
      if Assigned(psdAdmin) Then LocalFree(THandle(psdAdmin));
      if Assigned(psidAdmin) Then FreeSid(psidAdmin);
      if (ImpersonationToken <> 0) Then CloseHandle(ImpersonationToken);
      if (Token <> 0) Then CloseHandle(Token);
   end;
end;
end.
